using System;
using System.IO;
using System.Diagnostics;
using System.Reflection;
using System.Reflection.Emit;
using System.Threading;

namespace EmitCode
{
    /// <summary>
    /// Generate code for condition
    /// </summary>
    internal class CodeGenerator
    {
        internal Object Generate()            
        {
            ModuleBuilder module = EmitAssemblyModule();
            Type newCls = EmitClass(module);
            return module.Assembly.CreateInstance(newCls.Name);
        }


        internal CodeGenerator(Type elemType, Type expressionInterface, String condition) 
        {
            this.condition = condition;
            this.elemType = elemType;
            this.expressionInterface = expressionInterface;
        }

        internal enum TokenType 
        {
            tknEof,
            tknIdent,
            tknLpar,
            tknRpar,
            tknDot,
            tknIconst,
            tknLconst,
            tknSconst,
            tknFconst,
            tknComma,
            tknQuestion,
            tknColon,
            tknAdd,
            tknSub,
            tknMul,
            tknDiv,
            tknMod,
            tknShl,
            tknShr,
            tknXor,
            tknOr,
            tknAnd,
            tknLAnd,
            tknLOr,
            tknCom,
            tknNot,
            tknNull,
            tknNeg,
            tknEq,
            tknNe,
            tknGt,
            tknGe,
            tknLt,
            tknLe
        };
        

        internal TokenType Scan() 
        {
            int p = pos;
            int eol = condition.Length;
            char ch = '\0';
            int i;
            while (p < eol && Char.IsWhiteSpace(ch = condition[p])) 
            { 
                p += 1;
            }
            if (p == eol) 
            { 
                return TokenType.tknEof;
            }
            pos = ++p;
            switch (ch) 
            { 
                case '+':
                    return TokenType.tknAdd;
                case '-':
                    return TokenType.tknSub;
                case '*':
                    return TokenType.tknMul;
                case '/':
                    return TokenType.tknDiv;
                case '%':
                    return TokenType.tknMod;
                case '.':
                    return TokenType.tknDot;
                case '(':
                    return TokenType.tknLpar;
                case ')':
                    return TokenType.tknRpar;
                case ',':
                    return TokenType.tknComma;
                case '?':
                    return TokenType.tknQuestion;
                case ':':
                    return TokenType.tknColon;
                case '<':
                    if (p < eol) 
                    { 
                        if (condition[p] == '=') 
                        { 
                            pos += 1;
                            return TokenType.tknLe;
                        } 
                        if (condition[p] == '<') 
                        { 
                            pos += 1;
                            return TokenType.tknShl;
                        }
                    }
                    return TokenType.tknLt;
                case '>':
                    if (p < eol) 
                    { 
                        if (condition[p] == '=') 
                        { 
                            pos += 1;
                            return TokenType.tknGe;
                        } 
                        if (condition[p] == '>') 
                        { 
                            pos += 1;
                            return TokenType.tknShr;
                        }
                    }
                    return TokenType.tknGt;
                case '=':
                    if (p == eol || condition[p] != '=') 
                    { 
                        throw new CompileError("'=' expected", p-1);
                    } 
                    pos += 1;
                    return TokenType.tknEq;
                case '!':
                    if (p == eol || condition[p] != '=') 
                    { 
                        throw new CompileError("Invalid token '!'", p-1);
                    } 
                    pos += 1;
                    return TokenType.tknNe;
                case '|':
                    if (p < eol && condition[p] == '|') 
                    {
                        pos += 1;
                        return TokenType.tknLOr;
                    }
                    return TokenType.tknOr;
                case '&':
                    if (p < eol && condition[p] == '&') 
                    {
                        pos += 1;
                        return TokenType.tknLAnd;
                    }
                    return TokenType.tknAnd;
                case '\"':
                    i = 0; 
                    while (p < eol && condition[p] != '\"') 
                    { 
                        if (condition[p] == '\\') 
                        {
                            p += 1;
                        }
                        i += 1;
                        p += 1;
                    }
                    if (p >= eol) 
                    { 
                        throw new CompileError("Unexpected end of string constant", p);
                    }
                    char[] buf = new char[i];
                    for (p = pos, i = 0; condition[p] != '\"'; p++, i++) 
                    {
                        if (condition[p] == '\\') 
                        {
                            p += 1;
                        }
                        buf[i] = condition[p];
                    }
                    svalue = new String(buf, 0, i);
                    pos = p+1;
                    return TokenType.tknSconst;
                case '0': case '1': case '2': case '3': case '4': 
                case '5': case '6': case '7': case '8': case '9':
                    i = p - 1;
                    while (p < eol && Char.IsNumber(ch = condition[p])) 
                    { 
                        p += 1;
                    }
                    if (ch == '.' || ch == 'e' || ch == 'E') 
                    { 
                        while (++p < eol && (Char.IsNumber(condition[p]) 
                            || condition[p] == 'e' || condition[p] == 'E' || condition[p] == '.' ||
                            ((ch == 'e' || ch == 'E') 
                            && (condition[p] == '-' || condition[p] == '+'))));
                        pos = p;
                        fvalue = Double.Parse(condition.Substring(i, p-i)); 
                        return TokenType.tknFconst;
                    } 
                    else if (ch == 'l' || ch == 'L') 
                    { 
                        pos = p;
                        lvalue = Int64.Parse(condition.Substring(i, p-i));
                        return TokenType.tknLconst;
                    } 
                    else 
                    {
                        pos = p;
                        ivalue = Int32.Parse(condition.Substring(i, p-i));
                        return TokenType.tknIconst;                    
                    }
                default:
                    if (Char.IsLetter(ch) || ch == '$' || ch == '_') 
                    { 
                        i = p-1;
                        while (p < eol && (Char.IsLetterOrDigit(ch = condition[p]) 
                            || ch == '$' || ch == '_'))
                        {
                            p += 1;
                        }
                        pos = p;
                        ident = condition.Substring(i, p-i);
                        return ident.Equals("null") ? TokenType.tknNull : TokenType.tknIdent;
                    } 
                    else 
                    { 
                        throw new CompileError("Invalid symbol: " + ch, p-1);
                    }
            }
        }
        
        internal Type Condition() 
        {
            int p = pos;
            Type type = Disjunction();
            if (lex == TokenType.tknQuestion) 
            {
                if (type != typeof(bool)) 
                {
                    throw new CompileError("Boolean expression expected", p);
                }
                Label then = il.DefineLabel();
                Label endif = il.DefineLabel();
                il.Emit(OpCodes.Brfalse_S, then);
                Type ifTrue = Condition();
                if (lex != TokenType.tknColon) 
                {
                    throw new CompileError("':' expected", pos);
                }
                p = pos;
                il.Emit(OpCodes.Br_S, endif);
                il.MarkLabel(then);
                Type ifFalse = Condition();
                if (ifFalse != ifTrue) 
                { 
                    throw new CompileError("Types of alternatives of '?' expression should be the same", p);
                }
                il.MarkLabel(endif);
                type = ifTrue;
            }
            return type;
        }

        internal Type Disjunction() 
        {
            Type left = Conjunction();
            while (true) 
            {
                if (lex == TokenType.tknLOr) 
                {
                    if (left == typeof(bool)) 
                    {
                        LocalBuilder tmp = il.DeclareLocal(typeof(bool));
                        il.Emit(OpCodes.Stloc, tmp);
                        il.Emit(OpCodes.Ldloc, tmp);
                        Label ifTrue = il.DefineLabel();
                        il.Emit(OpCodes.Brtrue_S, ifTrue);
                        Type right = Comparison();
                        if (right == typeof(bool)) 
                        { 
                            il.Emit(OpCodes.Stloc, tmp);
                            il.MarkLabel(ifTrue);
                            il.Emit(OpCodes.Ldloc, tmp);
                            continue;
                        }
                        /*
                        il.Emit(OpCodes.Dup);
                        Label ifTrue = il.DefineLabel();
                        il.Emit(OpCodes.Brtrue_S, ifTrue);
                        Type right = Conjunction();
                        if (right == typeof(bool)) 
                        { 
                            il.MarkLabel(ifTrue);
                            continue;
                        }
                        */
                    } 
                    throw new CompileError("Oparator || can be applicaent only to boolean operands", pos);
                } 
                else if (lex == TokenType.tknOr) 
                {
                    Type right = Conjunction(); 
                    if (left != right || (left != typeof(int) && left != typeof(long))) 
                    { 
                        throw new CompileError("Oparator | can be applicaent only to integer operands of the same type", pos);
                    }
                    il.Emit(OpCodes.Or);
                } 
                else 
                {
                    break;
                }
            }
            return left;
        }     

        internal Type Conjunction() 
        {
            Type left = Comparison();
            while (true) 
            {
                if (lex == TokenType.tknLAnd) 
                {
                    if (left == typeof(bool)) 
                    {
                        LocalBuilder tmp = il.DeclareLocal(typeof(bool));
                        il.Emit(OpCodes.Stloc, tmp);
                        il.Emit(OpCodes.Ldloc, tmp);
                        Label ifFalse = il.DefineLabel();
                        il.Emit(OpCodes.Brfalse_S, ifFalse);
                        Type right = Comparison();
                        if (right == typeof(bool)) 
                        { 
                            il.Emit(OpCodes.Stloc, tmp);
                            il.MarkLabel(ifFalse);
                            il.Emit(OpCodes.Ldloc, tmp);
                            continue;
                        }

                        /*
                        il.Emit(OpCodes.Dup);
                        Label ifFalse = il.DefineLabel();
                        il.Emit(OpCodes.Brfalse_S, ifFalse);
                        Type right = Comparison();
                        if (right == typeof(bool)) 
                        { 
                            il.MarkLabel(ifFalse);
                            continue;
                        }
                        */
                    } 
                    throw new CompileError("Oparator && can be applicant only to boolean operands", pos);
                } 
                else if (lex == TokenType.tknAnd) 
                {
                    Type right = Comparison(); 
                    if (left != right || (left != typeof(int) && left != typeof(long))) 
                    { 
                        throw new CompileError("Oparator & can be applicent only to integer operands of the same type", pos);
                    }
                    il.Emit(OpCodes.And);
                } 
                else 
                {
                    break;
                }
            }
            return left;
        }     

        internal Type Comparison() 
        {
            Type left = Shift();
            TokenType tkn = lex;
            while (tkn == TokenType.tknEq || tkn == TokenType.tknNe || tkn == TokenType.tknGt 
                || tkn == TokenType.tknGe || tkn == TokenType.tknLe || tkn == TokenType.tknLt)
            {
                int p = pos;
                Type right = Shift();
                if (left != right) 
                { 
                    throw new CompileError("Comparison operands should have the same type", p);
                }
                if (left == typeof(bool) && tkn != TokenType.tknEq && tkn != TokenType.tknNe) 
                {
                    throw new CompileError("Boolean operands can be compared only for equality", p);
                }

                switch (tkn) 
                {
                    case TokenType.tknNe:
                        il.Emit(OpCodes.Ceq);
                        il.Emit(OpCodes.Ldc_I4_0);
                        il.Emit(OpCodes.Ceq);
                        continue;
                    case TokenType.tknEq:
                        il.Emit(OpCodes.Ceq);
                        break;
                    case TokenType.tknLt:
                        il.Emit(OpCodes.Clt);
                        break;
                    case TokenType.tknLe:
                        il.Emit(OpCodes.Cgt);
                        il.Emit(OpCodes.Ldc_I4_0);
                        il.Emit(OpCodes.Ceq);
                        break;
                    case TokenType.tknGt:
                        il.Emit(OpCodes.Cgt);
                        break;
                    case TokenType.tknGe:
                        il.Emit(OpCodes.Clt);
                        il.Emit(OpCodes.Ldc_I4_0);
                        il.Emit(OpCodes.Ceq);
                        break;
                }
                left = typeof(bool);
                tkn = lex;
            }
            return left;
        }
         
        internal Type Shift() 
        {
            Type left = Addition();
            TokenType tkn = lex;
            while (tkn == TokenType.tknShr || tkn == TokenType.tknShl) 
            {
                int p = pos;
                Type right = Addition();
                if ((left != typeof(int) && left != typeof(long))
                    || (right != typeof(int) && right != typeof(long)))
                {
                    throw new CompileError("Shift can be applient only to integer operands", p);
                }
                il.Emit(lex == TokenType.tknShr ? OpCodes.Shr : OpCodes.Shl);
                tkn = lex;
            }
            return left;
        }

        internal Type Addition()
        {
            Type left = Multiplication();
            TokenType tkn = lex;
            while (tkn == TokenType.tknAdd || tkn == TokenType.tknSub) 
            {
                int p = pos;
                Type right = Multiplication();
                if (left != right || (left != typeof(int) && left != typeof(long) && left != typeof(double)))
                {
                    throw new CompileError("Only integer or real operands of the same type can be added/subtracted", p);
                }
                il.Emit(tkn == TokenType.tknAdd ? OpCodes.Add : OpCodes.Sub);
                tkn = lex;
            }
            return left;
        }

        internal Type Multiplication()
        {
            Type left = Unary();
            TokenType tkn = lex;
            while (tkn == TokenType.tknMul || tkn == TokenType.tknDiv || tkn == TokenType.tknMod) 
            {
                int p = pos;
                Type right = Unary();
                if (left != right || (left != typeof(int) && left != typeof(long) && left != typeof(double)))
                {
                    throw new CompileError("Only integer or real operands of the same type can be multiplied/divided", p);
                }
                il.Emit(tkn == TokenType.tknMul ? OpCodes.Mul : tkn == TokenType.tknDiv ? OpCodes.Div : OpCodes.Rem);
                tkn = lex;
            }
            return left;
        }

        internal static Type GetStackType(Type type) 
        {
            if (type == typeof(byte) || type == typeof(char) || type == typeof(short) 
                || type == typeof(int) || type == typeof(sbyte) || type == typeof(ushort) 
                || type == typeof(uint))
            {
                return typeof(int);
            } 
            else if (type == typeof(long) || type == typeof(ulong)) 
            {
                return typeof(long);
            } 
            else if (type == typeof(float) || type == typeof(double)) 
            {
                return typeof(double);
            }
            return type;
        }


        internal enum AccessType 
        {
            This,
            Static,
            Instance
        };

       internal void Expect(TokenType tkn) 
        { 
            int p = pos;
            if (Scan() != tkn) 
            { 
                throw new CompileError("Unexpected token", p);
            }
        }


        internal Type MemberAccess(Type type, String ident, AccessType access)
        {
            int p = pos;
            int i, j;
            MemberInfo[] mi = type.GetMember(ident);
            MethodInfo mth;
            if (mi == null || mi.Length == 0) 
            { 
                if (access == AccessType.This) 
                {
                    String className = ident;
                    while (Scan() == TokenType.tknDot) 
                    {
                        p = pos;
                        Expect(TokenType.tknIdent);
                        className += ident;
                        Type cls = Type.GetType(className);
                        if (cls != null) 
                        {
                            Expect(TokenType.tknDot);
                            Expect(TokenType.tknIdent);
                            return MemberAccess(cls, ident, AccessType.Static);
                        }
                    }
                }
                throw new CompileError("No such member '" + ident + "' in class " + type, p);
            }
            p = pos;
            lex = Scan();
            if (lex == TokenType.tknLpar) 
            {
                bool isStaticMethod = false;
                bool isInstanceMethod = false; 
                for (i = 0; i < mi.Length; i++) 
                {
                    if ((mth = mi[i] as MethodInfo) != null) 
                    {
                        if (access == AccessType.Static) 
                        { 
                            if (mth.IsStatic) 
                            {
                                // exclude other non-static alternatives from 
                                for (j = 0; j < mi.Length; j++) 
                                { 
                                    MethodBase m = mi[j] as MethodInfo;
                                    if (!m.IsStatic) 
                                    { 
                                        mi[j] = null;
                                    }
                                }
                                isStaticMethod = true;
                                break;
                            }
                        }
                        else 
                        {
                            if (!mth.IsStatic) 
                            {
                                // exclude other static alternatives from 
                                for (j = 0; j < mi.Length; j++) 
                                { 
                                    MethodInfo m = mi[j] as MethodInfo;
                                    if (m.IsStatic) 
                                    { 
                                        mi[j] = null;
                                    }
                                }
                                isInstanceMethod = true;
                                break;
                            }
                        }
                    }
                }
                if (!isStaticMethod && !isInstanceMethod) 
                { 
                    for (j = 0; j < mi.Length; j++) 
                    { 
                        if ((mth = mi[j] as MethodInfo) != null) {
                            if (mth.IsStatic) 
                            {
                                if (isInstanceMethod) 
                                { 
                                    new CompileError("Overloading can not be resolved for method " + mth.Name, p);
                                }
                                isStaticMethod = true;
                            } 
                            else 
                            {
                                if (isStaticMethod) 
                                { 
                                    new CompileError("Overloading can not be resolved for method " + mth.Name, p);
                                }
                                isStaticMethod = true;
                            }
                        }
                    }
                }
                if (isStaticMethod) 
                {
                    if (access == AccessType.Instance) 
                    {
                        il.Emit(OpCodes.Pop);
                    }
                } 
                else 
                {
                    if (access == AccessType.This) 
                    {
                        il.Emit(OpCodes.Ldloc_0, local);
                    }
                }
           
                int n = 0;
                do 
                {
                    p = pos;
                    Type actualParamType = Condition();
                    Type bestMatchType = null;
                    for (i = 0; i < mi.Length; i++) 
                    { 
                        if ((mth = mi[i] as MethodInfo) != null) 
                        { 
                            ParameterInfo[] methodParams = mth.GetParameters();
                            if (methodParams.Length <= n) 
                            {
                                mi[i] = null;
                                continue;
                            }
                            Type formalParamType = methodParams[n].ParameterType;
                            if (formalParamType == actualParamType) 
                            {
                                bestMatchType = GetStackType(formalParamType);
                                for (j = 0; j < mi.Length; j++) 
                                { 
                                    if ((mth = mi[j] as MethodInfo) != null) 
                                    { 
                                        methodParams = mth.GetParameters();
                                        if (methodParams.Length <= n || methodParams[n].ParameterType != actualParamType) 
                                        { 
                                            mi[j] = null;
                                        }
                                    }
                                }
                                break;
                            }
                            if (!formalParamType.IsAssignableFrom(actualParamType)) 
                            {
                                mi[i] = null;
                            } 
                            else if (bestMatchType == null) 
                            { 
                                bestMatchType = formalParamType;
                            }
                            else if (bestMatchType != formalParamType) 
                            {
                                 if (bestMatchType.IsAssignableFrom(formalParamType)) 
                                 {
                                     for (j = i; --j >= 0;) 
                                     { 
                                         if ((mth = mi[j] as MethodInfo) != null) 
                                         { 
                                             if (mth.GetParameters()[n].ParameterType == bestMatchType) 
                                             { 
                                                 mi[j] = null;
                                             }
                                         }
                                     }
                                     bestMatchType = formalParamType;
                                 } 
                                 else if (formalParamType.IsAssignableFrom(bestMatchType)) 
                                 {
                                     mi[i] = null;
                                 }
                            }
                        }
                    }
                    if (bestMatchType == null) 
                    { 
                        throw new CompileError("Failed to find best matched overloading method", p);
                    }
                    bestMatchType = GetStackType(bestMatchType);
                    if (actualParamType == typeof(int) && bestMatchType == typeof(long)) 
                    {
                        il.Emit(OpCodes.Conv_I8);
                    } 
                    else if (actualParamType == typeof(long) && bestMatchType == typeof(int)) 
                    {
                        il.Emit(OpCodes.Conv_I4);
                    }
                    n += 1;
                } while (lex == TokenType.tknComma);

                if (lex != TokenType.tknRpar) 
                { 
                    throw new CompileError("')' expected", pos);
                }
                mth = null;
                for (i = 0; i < mi.Length; i++) 
                { 
                    MethodInfo m = mi[i] as MethodInfo;
                    if (m != null) 
                    { 
                        if (mth != null) 
                        { 
                            throw new CompileError("Ambigious overloading for method " + mth.Name, p);
                        }
                        mth = m;
                    }
                }
                if (mth == null) 
                { 
                    throw new CompileError("Icompatible types for method parameters", p);
                }
                il.Emit(mth.IsVirtual ? OpCodes.Callvirt : OpCodes.Call, mth);
                type = GetStackType(mth.ReturnType);
                lex = Scan();
            }
            else 
            {
                for (i = 0; i < mi.Length; i++) 
                { 
                    FieldInfo fld = mi[i] as FieldInfo;
                    if (fld != null) 
                    {
                        if (fld.IsStatic) 
                        {
                            if (access == AccessType.Instance) 
                            {
                                il.Emit(OpCodes.Pop);
                            }
                            il.Emit(OpCodes.Ldsfld);
                        } 
                        else
                        {
                            if (access == AccessType.This) 
                            {
                                il.Emit(OpCodes.Ldloc_0, local);
                            }
                            il.Emit(OpCodes.Ldfld, fld);
                        }
                        type = GetStackType(fld.FieldType);      
                    } 
                    else 
                    {
                        PropertyInfo prop = mi[i] as PropertyInfo;
                        if (prop == null) 
                        { 
                            throw new CompileError("Member '" + mi[i].Name 
                                + "' is neither FieldAccessException, neither property", p);
                        }
                        mth = prop.GetGetMethod();
                        if (mth == null) 
                        { 
                            throw new CompileError("Property '" + mi[i].Name + "' has no getter", p);
                        }
                        if (mth.IsStatic) 
                        {
                            if (access == AccessType.Instance) 
                            {
                                il.Emit(OpCodes.Pop);
                            }
                        } 
                        else 
                        {
                            if (access == AccessType.This) 
                            {
                                il.Emit(OpCodes.Ldloc_0, local);
                            }
                        } 
                        il.Emit(mth.IsVirtual ? OpCodes.Callvirt : OpCodes.Call, mth);
                        type = GetStackType(mth.ReturnType);
                    }
                }
            }
            return type;
        }

        internal Type Unary() 
        {
            int p = pos;
            Type type;
            TokenType tkn;
            switch (tkn = Scan()) 
            {
                case TokenType.tknNot:
                case TokenType.tknCom:
                case TokenType.tknAdd:
                case TokenType.tknSub:
                    p = pos;
                    type = Unary();
                    switch (tkn) 
                    {
                        case TokenType.tknNot:
                            if (type != typeof(bool)) 
                            {
                                throw new CompileError("Operator ! can be appliead only to operand of boolean type", p);
                            }
                            il.Emit(OpCodes.Not);               
                            break;
                        case TokenType.tknCom:
                            if (type != typeof(int) && type != typeof(long)) 
                            {
                                throw new CompileError("Operator ~ can be appliead only to operand of integer type", p);
                            }
                            il.Emit(OpCodes.Not);               
                            break;
                        case TokenType.tknSub:
                            if (type != typeof(int) && type != typeof(long) && type != typeof(double)) 
                            {
                                throw new CompileError("Operator - can be appliead only to operand of integer or real type", p);
                            }
                            il.Emit(OpCodes.Neg);  
                            break;
                    }
                    return type;
                case TokenType.tknNull:
                    il.Emit(OpCodes.Ldnull);
                    type = typeof(Object);
                    lex = Scan();
                    break;
                case TokenType.tknLpar:
                    type = Condition();
                    if (lex != TokenType.tknRpar) 
                    {
                        throw new CompileError("')' expected", pos);
                    }
                    lex = Scan();
                    break;
                case TokenType.tknIconst:
                    il.Emit(OpCodes.Ldc_I4, ivalue);
                    type = typeof(int);
                    lex = Scan();
                    break;
                case TokenType.tknLconst:
                    il.Emit(OpCodes.Ldc_I8, lvalue);
                    type = typeof(long);
                    lex = Scan();
                    break;
                case TokenType.tknFconst:
                    il.Emit(OpCodes.Ldc_R8, fvalue);
                    type = typeof(double);
                    lex = Scan();
                    break;
                case TokenType.tknSconst:
                    il.Emit(OpCodes.Ldstr, svalue);
                    type = typeof(String);
                    lex = Scan();
                    break;
                case TokenType.tknIdent:
                    type = MemberAccess(elemType, ident, AccessType.This);
                    break;
                default:
                    throw new CompileError("Unexpected token", p);
            }
            while (lex == TokenType.tknDot) 
            {
                Expect(TokenType.tknIdent);
                type = MemberAccess(type, ident, AccessType.Instance);
            }
            return type;
        }

        ///<summary>
        ///Create an assembly that will evaluate the polynomial.
        ///</summary>
        private ModuleBuilder EmitAssemblyModule()
        {
            if (dynamicModule == null)
            {
                //
                //Create an assembly name
                //
                AssemblyName assemblyName = new AssemblyName();
                assemblyName.Name = "DynamicFilterAssembly";
                //
                //Create a new assembly with one module
                //
                AssemblyBuilder assembly = Thread.GetDomain().DefineDynamicAssembly(assemblyName, AssemblyBuilderAccess.Run);
                dynamicModule = assembly.DefineDynamicModule("EvaluateExpression");
            }
            return dynamicModule;
        }

        private Type EmitClass(ModuleBuilder module)
        {
            //
            //Define a public class named "ExpressionN" in the assembly.
            //
            counter += 1;
            String generatedClassName = "GeneratedClass" + counter;
            TypeBuilder myType = module.DefineType(generatedClassName, TypeAttributes.Public);
            //
            //Mark the class as implementing ICondition.
            //
            myType.AddInterfaceImplementation(expressionInterface);
            //Add a constructor
            ConstructorBuilder constructor =
                myType.DefineDefaultConstructor(MethodAttributes.Public);

            MethodInfo evalInterface = expressionInterface.GetMethod("Evaluate");
            Type returnType = evalInterface.ReturnType;
            ParameterInfo[] methodParams = evalInterface.GetParameters();
            Type[] paramTypes = new Type[methodParams.Length];
            for (int i = 0; i < methodParams.Length; i++) 
            {
                paramTypes[i] = methodParams[i].ParameterType;
            }
            MethodBuilder evalMethod =
                myType.DefineMethod("Evaluate",
                                    MethodAttributes.Public | MethodAttributes.Virtual,
                                    returnType,
                                    paramTypes);
            //
            //From the method,get an ILGenerator.This is used to
            //emit the IL that we want.
            //
            
            il = evalMethod.GetILGenerator();
            il.Emit(OpCodes.Ldarg_1);
            il.Emit(OpCodes.Castclass, elemType);
            local = il.DeclareLocal(elemType);
            il.Emit(OpCodes.Stloc_0, local);

            Type expressionType = Condition();
            if (returnType == typeof(bool)) 
            {
                if (expressionType != typeof(bool)) 
                {
                    throw new CompileError("Condition expression should be of boolean type", 0);
                } 
            }
            else if (returnType != expressionType) 
            {
                if (returnType == typeof(Object)) 
                {
                    if (expressionType == typeof(int) || expressionType == typeof(long) 
                        || expressionType == typeof(bool) || expressionType == typeof(double))
                    {
                        il.Emit(OpCodes.Box, expressionType);
                    }
                } 
                else 
                {
                    throw new CompileError("Invalid type of expression: " + returnType + " expected", 0);
                }
            }
            if (lex != TokenType.tknEof) 
            {
                throw new CompileError("Invalid expression", pos);
            }
            il.Emit(OpCodes.Ret);
            myType.DefineMethodOverride(evalMethod, evalInterface);
            myType.CreateType();
            return myType;
        }

        private String        condition;
        private Type          elemType;
        private Type          expressionInterface;
        private int           pos;
        private String        ident;
        private String        svalue;
        private int           ivalue;
        private long          lvalue;
        private double        fvalue;
        private TokenType     lex;
        private ILGenerator   il;
        private LocalBuilder  local;
        private static ModuleBuilder dynamicModule;
        private static int    counter = 0;
    }
}
